#include <simpleDCD.h>
#include <data.h>
#include <zlib.h>
#include <lbfgssolver.h>
#include <set>
#include <cmath>
#include <cassert>
#include <cstdlib>
#include <cstring>

simpleDCD::simpleDCD ( const char *fname ) {
  dataptr = NULL, label = -1, w = NULL, bias = 0;
  lambda = 0;
  P = 0;
  readBinary( fname );
}


simpleDCD::simpleDCD ( Data* _dataptr , double _lambda , int _bias ) {
  dataptr = _dataptr;
  lambda = _lambda;
  label = -1;
  bias = _bias;
  P = 0;

  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
  }
  P += bias;

  w = (double*) malloc(sizeof(double)*P);
  memset( w , 0 , sizeof(double)*P );

}

#define INF (1<<30)

double simpleDCD::solve ( int _label , double eps ) {

  assert( dataptr );
  label = _label;

  double F = 0;

  int N = dataptr->ndocs;

  int *index = new int[N], *target = new int[N],max_iter = 1000, active_size = N;
  double *QD = new double[N],*alpha = new double[N],PG,PGmax_old = INF, PGmin_old = -INF;
  double C = lambda;

  // Init

  for ( int i = 0; i < N; ++i ) {

    alpha[i] = 0, QD[i] = 0, index[i] = i;

    FLOOP ( f , dataptr->docs[i] ) QD[i] += (*f.val)*(*f.val);

    if( bias ) QD[i] += 1; // 1 for the Bias term

    bool found = false;

    LLOOP( l , dataptr->docs[i] ) found |= ( (*l) == label );

    target[i] = ( found == true ? +1 : -1 );
  }


  // Perform Iterations

  for( int iter = 0; iter < max_iter;  ){

    double PGmax_new = -INF, PGmin_new = INF;

    for ( int i = 0; i < active_size; ++i ) std::swap( index[i] , index[ i+rand()%(active_size-i) ] );

    for ( int s = 0; s < active_size; ++s ) {

      int i = index[s];

      double G = 0;

      FLOOP ( f , dataptr->docs[i] ) G += w[ (*f.id) ] * (*f.val);

      if( bias ) G += w[P-1] * 1; // adding the bias term

      G = G * target[i] - 1;

      PG = 0;

      if( alpha[i] == 0 ) {
	if( G > PGmax_old ) {
	  active_size--;
	  std::swap( index[s] , index[active_size] );
	  s--;
	  continue;
	}
	else if( G < 0 ) PG = G;
      } 
      else if( alpha[i] == C ) {
	if( G < PGmin_old ){
	  active_size--;
	  std::swap(index[s],index[active_size]);
	  s--;
	  continue;
	}
	else if( G > 0 ) PG = G;
      } 
      else PG = G;

      PGmax_new = std::max( PGmax_new , PG );
      PGmin_new = std::min( PGmin_new , PG );

      if( fabs(PG) > 1.0e-12 ) {
	double alpha_old = alpha[i];
	alpha[i] = std::min( std::max( alpha[i] - G/QD[i] , 0.0 ) , C );
	double d = (alpha[i] - alpha_old) * target[i];

	FLOOP(f,dataptr->docs[i]) w[ *f.id ] += d * (*f.val);
	if( bias ) w[P-1] += d * 1; // For the bias term
      }
    }
    iter++;
    if( PGmax_new - PGmin_new <= eps ) {

      if( active_size == N ) break;
      else {
	active_size = N;
	PGmax_old = INF;
	PGmin_old = -INF;
	continue;
      }
    }

    PGmax_old = PGmax_new;
    PGmin_old = PGmin_new;

    if( PGmax_old <= 0 ) PGmax_old = INF;
    if( PGmin_old >= 0 ) PGmin_old = -INF;
  }

  F = 0;
  for ( int i = 0; i < P; ++i ) {
    F += .5*w[i]*w[i];
  }

  for ( int i = 0; i < N; ++i ) {
    double score = classify( &dataptr->docs[i] );
    score = std::max( 1 - target[i]*score , 0.0 );
    F += C*score;
  }

  delete[] index;
  delete[] target;
  delete [] QD;
  delete[] alpha;


  return F;
}



double simpleDCD::classify ( const Doc * doc ) {
  double sc = bias ? w[P-1] : 0;
  FLOOP ( f , *doc ) if ( *f.id < P-bias ) sc += w[*f.id] * (*f.val);
  return sc;
}


void simpleDCD::writeBinary ( const char *fname ) {
  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  int K = 1;
  gzwrite( f , &K , sizeof(int) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , &bias , sizeof(int) );
  gzwrite( f , &label  , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);
}


void simpleDCD::readBinary ( const char *fname ) {
  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }
  int K = 1;
  gzread( f , &K , sizeof(int) );
  gzread( f , &P , sizeof(int) );
  gzread( f , &bias , sizeof(int) );
  gzread( f , &label , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  //  printf("{ %d %d %d %d %lf}\n", K , P , bias , label , w[P-1] );
}


simpleDCD::~simpleDCD (){
  free(w);
}

