#include <simpleGBLR.h>
#include <cstdlib>
#include <zlib.h>
#include <cstring>
#include <cmath>
#include <vector>
#include <lbfgssolver.h>
#include <cassert>
#include <simpleBLR.h>
#include <sys/time.h>

simpleGBLR::simpleGBLR ( int _label , const char *fname , HInfo * _hinfo ) {
  dataptr = NULL, label = _label, mean = NULL , w = NULL, bias = 0, hinfo = _hinfo;
  lambda = 0;
  P = 0;
  readBinary( label , fname );
}


simpleGBLR::simpleGBLR ( Data* _dataptr , double _lambda , HInfo * _hinfo, int _bias ) {
  dataptr = _dataptr;
  hinfo = _hinfo;
  lambda = _lambda;
  mean = NULL;
  label = -1;
  bias = _bias;
  P = 0;

  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
  }
  P += bias;

  w = (double*) malloc(sizeof(double)*P);
  memset( w , 0 , sizeof(double)*P );

}


double * simpleGBLR::getMeans ( const char *path ) {

  mean = (double *) malloc(sizeof(double)*P);
  memset( mean , 0 , sizeof(double)*P );
  char fname[512];

  std::vector<int> nbs = hinfo->getNeighbors( label );

  for ( int i = 0; i < nbs.size(); ++i ) {
    sprintf(fname,"%s/%d.gz", path , nbs[i] );
    if ( fexist(fname) ) {
      simpleGBLR N ( nbs[i] , fname , hinfo );
      assert( N.P == P );
      for ( int j = 0; j < N.P-N.bias; ++j )
	mean[j] += N.w[j];
    }
  }

  for ( int i = 0; i < P; ++i ) 
    mean[i] /= ( nbs.size() );
  return mean;
}


double simpleGBLR::solve ( const char *path , int _label , double eps ) {
  label = _label;

  mean = getMeans( path );
  int npos = 0, nneg = 0;
  int K = hinfo->getNeighbors( label ).size();

  DLOOP ( doc , *dataptr ) {
    int found = 0;
    LLOOP ( l , *doc ) found |= ( *l == label );
    npos += found;
    nneg += !found;
  }
  if ( npos == 0 ) {
    // Copy the mean into the current weight-vector
    memcpy ( w , mean , sizeof(double)*P );
    free(mean);
    return 0;
  }


  //==========================================================================
  assert( dataptr );
  lambda *= K;

  LBFGSsolver* solver = new LBFGSsolver( w , P , this , std::min(1e-6,lambda/100) ) ;
  double F = solver->objective();
  delete solver;

  lambda /= K;
  free(mean) , mean = NULL;

  return F;
}

#define INF (1<<30)
#define EPS (1e-50)

double simpleGBLR::fg ( const double *x , double *g , int n ) {

  double F = 0;
  g[n-1] = 0;
  for ( int i = 0; i < n-bias; ++i ) 
    g[i] = lambda * (x[i]-mean[i]) ,
      F += lambda/2 * (x[i]-mean[i])*(x[i]-mean[i]);


  DLOOP ( doc , *dataptr ) {
    int y = -1;
    LLOOP ( l , *doc ) y = ( *l == label ? +1: -1 );

    double wx = bias ? x[n-1] : 0;
    FLOOP ( f , *doc ) wx += x[*f.id] * (*f.val);
    double sc = exp(y*wx);
    if ( isinf(sc) ) sc = INF;


    double pr = 1 / ( 1 + sc );
    FLOOP ( f , *doc ) g[*f.id] += -y * pr * (*f.val);
    if ( bias ) g[n-1] += -y*pr;
    
    double add = exp(-y*wx);
    if ( isinf(add) )
      add = INF;
    else
      add = log(1+add);
    F += add;
  }
  return F;
}


double simpleGBLR::classify ( const Doc * doc ) {
  double sc = bias ? w[P-1] : 0;
  FLOOP ( f , *doc ) if ( *f.id < P-bias ) sc += w[*f.id] * (*f.val);
  return sc;
}


void simpleGBLR::writeBinary ( const char *fname ) {
  pthread_mutex_lock( &filelocks[label] );

  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  int K = 1;
  gzwrite( f , &K , sizeof(int) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , &bias , sizeof(int) );
  gzwrite( f , &label  , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);

  pthread_mutex_unlock( &filelocks[label] );
}


void simpleGBLR::readBinary ( int lockl , const char *fname ) {
  pthread_mutex_lock( &filelocks[lockl] );

  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  int K = 1;
  gzread( f , &K , sizeof(int) );
  gzread( f , &P , sizeof(int) );
  gzread( f , &bias , sizeof(int) );
  gzread( f , &label , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  pthread_mutex_unlock( &filelocks[lockl] );
}


simpleGBLR::~simpleGBLR (){
  free(w);
}

