#include <simpleGDCD.h>
#include <cstdlib>
#include <zlib.h>
#include <cstring>
#include <cmath>
#include <vector>
#include <lbfgssolver.h>
#include <cassert>
#include <simpleBLR.h>
#include <sys/time.h>

simpleGDCD::simpleGDCD ( int _label , const char *fname , HInfo * _hinfo ) {
  dataptr = NULL, label = _label, w = NULL, bias = 0, hinfo = _hinfo;
  lambda = 0;
  P = 0;
  readBinary( label , fname );
}


simpleGDCD::simpleGDCD ( Data* _dataptr , double _lambda , HInfo * _hinfo, int _bias ) {
  dataptr = _dataptr;
  hinfo = _hinfo;
  lambda = _lambda;
  label = -1;
  bias = _bias;
  P = 0;
  w = NULL;

  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
  }
  P += bias;

  w = (double*) malloc(sizeof(double)*P);
  memset( w , 0 , sizeof(double)*P );
}



double * simpleGDCD::getMeans ( const char *path  ) {

  double *mean = (double *) malloc(sizeof(double)*P);
  memset( mean , 0 , sizeof(double)*P );
  char fname[512];

  std::vector<int> nbs = hinfo->getNeighbors( label );

  for ( int i = 0; i < nbs.size(); ++i ) {
    sprintf(fname,"%s/%d.gz", path , nbs[i] );
    if ( fexist(fname) ) {
      simpleGDCD N ( nbs[i] , fname , hinfo );
      if( N.P != P ) {
	printf("N.P = %d N.label = %d ... P = %d label = %d ", N.P , N.label , P , label );
      }
      assert( N.P == P );
      for ( int j = 0; j < N.P; ++j )
	mean[j] += N.w[j];
    }
  }

  for ( int i = 0; i < P; ++i ) 
    mean[i] /= ( nbs.size() );
  return mean;
}


double simpleGDCD::solve ( const char *path , int _label , double eps ) {
  label = _label;

  double *mean = getMeans( path );
  int npos = 0, nneg = 0;
  int K = hinfo->getNeighbors( label ).size();

  DLOOP ( doc , *dataptr ) {
    int found = 0;
    LLOOP ( l , *doc ) found |= ( *l == label );
    npos += found;
    nneg += !found;
  }
  if ( npos == 0 ) {
    // Copy the mean into the current weight-vector
    memcpy ( w , mean , sizeof(double)*P );
    free(mean);
    return 0;
  }


  //==========================================================================
  const int INF = (1<<30);

  assert( dataptr );
  double F = 0;

  int N = dataptr->ndocs;

  int *index = new int[N], *target = new int[N],max_iter = 1000, active_size = N;
  double *QD = new double[N],*alpha = new double[N],PG,PGmax_old = INF, PGmin_old = -INF;
  double *pscore = new double[N];
  double C = lambda / K;

  // Init
  for ( int i = 0; i < N; ++i ) {
    alpha[i] = 0, QD[i] = 0, index[i] = i, pscore[i] = 0;
    FLOOP ( f , dataptr->docs[i] ) QD[i] += (*f.val)*(*f.val);
    if( bias ) QD[i] += 1; // 1 for the Bias term
    bool found = false;
    LLOOP ( l , dataptr->docs[i] ) if ( *l == label ) found = true;
    target[i] = ( found == true ? +1 : -1 );

    FLOOP ( f , dataptr->docs[i] ) pscore[i] += (*f.val)*mean[*f.id];
    if( bias ) pscore[i] += mean[P-1];
    pscore[i] = target[i]*pscore[i]; 
  }

  assert( npos && nneg );

  // Perform Iterations

  for( int iter = 0; iter < max_iter;  ){
    double PGmax_new = -INF, PGmin_new = INF;
    for ( int i = 0; i < active_size; ++i ) std::swap( index[i] , index[ i+rand()%(active_size-i) ] );
    for ( int s = 0; s < active_size; ++s ) {
      int i = index[s];
      double G = 0;
      FLOOP ( f , dataptr->docs[i] ) G += w[ (*f.id) ] * (*f.val);
      if( bias ) G += w[P-1] * 1; // adding the bias term
      G = G * target[i] - (1-pscore[i]);
      PG = 0;

      if( alpha[i] == 0 ) {
	if( G > PGmax_old ) {
	  active_size--;
	  std::swap( index[s] , index[active_size] );
	  s--;
	  continue;
	}
	else if( G < 0 ) PG = G;
      } 
      else if( alpha[i] == C ) {
	if( G < PGmin_old ){
	  active_size--;
	  std::swap(index[s],index[active_size]);
	  s--;
	  continue;
	}
	else if( G > 0 ) PG = G;
      } 
      else PG = G;

      PGmax_new = std::max( PGmax_new , PG );
      PGmin_new = std::min( PGmin_new , PG );

      if( fabs(PG) > 1.0e-12 ) {
	double alpha_old = alpha[i];
	alpha[i] = std::min( std::max( alpha[i] - G/QD[i] , 0.0 ) , C );
	double d = (alpha[i] - alpha_old) * target[i];

	FLOOP(f,dataptr->docs[i]) w[ *f.id ] += d * (*f.val);
	if( bias ) w[P-1] += d * 1; // For the bias term
      }
    }
    iter++;
    if( PGmax_new - PGmin_new <= eps ) {
      if( active_size == N ) break;
      else {
	active_size = N;
	PGmax_old = INF;
	PGmin_old = -INF;
	continue;
      }
    }

    PGmax_old = PGmax_new;
    PGmin_old = PGmin_new;

    if( PGmax_old <= 0 ) PGmax_old = INF;
    if( PGmin_old >= 0 ) PGmin_old = -INF;
  }

  for ( int i = 0; i < P; ++i )
    w[i] += mean[i];

  F = 0;
  for ( int i = 0; i < P; ++i ) {
    F += .5*(w[i]-mean[i])*(w[i]-mean[i]);
  }

  for ( int i = 0; i < N; ++i ) {
    double score = classify( &dataptr->docs[i] );
    score = std::max( 1 - target[i]*score , 0.0 );
    F += C*score;
  }

  delete[] index;
  delete[] target;
  delete [] QD;
  delete[] alpha;
  delete[] pscore;

  free(mean),  mean = NULL;
  return F;
}


void simpleGDCD::print ( FILE * f) {

  int K = 1;
  fprintf(f,"{K:%d, P:%d, bias:%d, label:%d}\n", K , P , bias, label );
  fprintf(f,"{");
  for ( int i = 0; i < P; ++i ) 
    fprintf(f,"%d:%lf ", i , w[i] );
  fprintf(f,"}\n");

}




double simpleGDCD::classify ( const Doc * doc ) {
  double sc = bias ? w[P-1] : 0;
  FLOOP ( f , *doc ) if ( *f.id < P-bias ) sc += w[*f.id] * (*f.val);
  return sc;
}


void simpleGDCD::writeBinary ( const char *fname ) {
  pthread_mutex_lock( &filelocks[label] );

  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  int K = 1;
  gzwrite( f , &K , sizeof(int) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , &bias , sizeof(int) );
  gzwrite( f , &label  , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);

  pthread_mutex_unlock( &filelocks[label] );
}


void simpleGDCD::readBinary ( int lockl , const char *fname ) {
  pthread_mutex_lock( &filelocks[lockl] );

  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  int K = 1;
  gzread( f , &K , sizeof(int) );
  gzread( f , &P , sizeof(int) );
  gzread( f , &bias , sizeof(int) );
  gzread( f , &label , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  pthread_mutex_unlock( &filelocks[lockl] );
}


simpleGDCD::~simpleGDCD (){
  free(w);
}

