#include <simpleHBLR.h>
#include <cstdlib>
#include <zlib.h>
#include <cstring>
#include <cmath>
#include <vector>
#include <lbfgssolver.h>
#include <cassert>
#include <simpleBLR.h>
#include <sys/time.h>

simpleHBLR::simpleHBLR ( int _label , const char *fname , HInfo * _hinfo ) {
  dataptr = NULL, label = _label, w = NULL, wp = NULL, bias = 0, hinfo = _hinfo;
  lambda = 0;
  P = 0;
  readBinary( label , fname );
}


simpleHBLR::simpleHBLR ( Data* _dataptr , double _lambda , HInfo * _hinfo, int _bias ) {
  dataptr = _dataptr;
  hinfo = _hinfo;
  lambda = _lambda;
  label = -1;
  bias = _bias;
  wp = NULL;
  P = 0;

  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
  }
  P += bias;

  w = (double*) malloc(sizeof(double)*P);
  memset( w , 0 , sizeof(double)*P );

}


double simpleHBLR::solve ( const char *path , int _label , double eps ) {
  
  if ( hinfo->isLeaf( _label ) == 1) 
    return solveLeaf ( path , _label , eps );
  else
    return solveNonLeaf ( path , _label , eps );
}


double simpleHBLR::solveLeaf ( const char *path , int _label , double eps )  {
  assert( dataptr );
  label = _label;

  wp = (double*) realloc(wp,sizeof(double)*P);
  memset( wp , 0 , sizeof(double)*P);

  char fname[512];
  sprintf(fname,"%s/%d.gz", path , hinfo->Parent(label) );
  if ( fexist(fname) ) {
    simpleHBLR Par( hinfo->Parent(label) , fname , hinfo );
    assert( Par.bias == 0 );
    memcpy( wp , Par.w , sizeof(double)*Par.P);
  }

  LBFGSsolver* solver = new LBFGSsolver( w , P , this , std::min(1e-5,lambda*10) ) ;
  double F = solver->objective();
  delete solver;

  free(wp),  wp = NULL;
  
  return F;
}



double simpleHBLR::solveNonLeaf ( const char *path , int _label , double eps ) {
  label = _label;

  assert( bias == 0 );
  double F = 0;
  int cnt = 0;
  char fname[512];
  std::vector<int> avgs = hinfo->Children(label);

  for ( int i = 0; i < avgs.size(); ++i ) {
    sprintf(fname,"%s/%d.gz", path , avgs[i] );
    if ( fexist(fname) ) {
      simpleHBLR N ( avgs[i] , fname , hinfo );
      for ( int j = 0; j < N.P-N.bias; ++j )
	w[j] += N.w[j];
    }
    cnt++;
  }

  int par = hinfo->Parent(label);
  sprintf(fname,"%s/%d.gz", path , par );
  if ( fexist(fname) ) {
    simpleHBLR N ( par , fname , hinfo );
    for ( int j = 0; j < N.P; ++j ) 
      w[j] += N.w[j];
    cnt++;
    for ( int j = 0; j < P; ++j ) 
      w[j] /= cnt, F += lambda/2*(w[j]-N.w[j])*(w[j]-N.w[j]);    
  } 
  else {
    for ( int j = 0; j < P; ++j ) 
      w[j] /= cnt;
  }

  return F;
}



#define INF (1<<30)
#define EPS (1e-50)
/*
double simpleHBLR::fg ( const double *x , double *g , int n ) {

  double F = 0;
  g[n-1] = 0;
  for ( int i = 0; i < n-bias; ++i ) 
    g[i] = lambda * (x[i]-wp[i]) ,
      F += lambda/2 * (x[i]-wp[i])*(x[i]-wp[i]);


  DLOOP ( doc , *dataptr ) {
    double sc = bias ? x[n-1] : 0;
    FLOOP ( f , *doc ) sc += x[*f.id] * (*f.val);
    sc = isinf(sc=exp(sc)) ? INF: sc;

    double pr = sc / ( 1 + sc );
    int y = ( label == doc->labels[0] );
    FLOOP ( f , *doc ) g[*f.id] += -(y-pr) * (*f.val);
    if ( bias ) g[n-1] += -(y-pr);

    double add = 0;
    if ( y == 1 ) {
      if ( pr < EPS ) add = INF;
      else add = -log(pr);
    }
    else {
      if ( pr > 1- EPS ) add = INF;
      else add = -log(1-pr);
    }
    F += add;
  }
  return F;
}
*/


double simpleHBLR::fg ( const double *x , double * g , int n ) {
  double F = 0;
  g[n-1] = 0;
  for ( int i = 1; i < n-bias; ++i ) 
    g[i] = lambda * (x[i]-wp[i]),
      F += lambda/2 *(x[i]-wp[i])*(x[i]-wp[i]);

  DLOOP ( doc , *dataptr ) {
    int y = -1;
    LLOOP ( l , *doc ) if ( *l == label ) y = +1;
    double wx = bias ? x[n-1] : 0;
    FLOOP ( f , *doc ) wx += x[*f.id] * (*f.val);
    double sc = exp(y*wx);
    if ( isinf(sc) ) sc = INF;

    double pr = 1 / ( 1 + sc );
    FLOOP ( f , *doc ) g[*f.id] += -y * pr * (*f.val);
    if ( bias ) g[n-1] += -y*pr;
    
    double add = exp(-y*wx);
    if ( isinf(add) )
      add = INF;
    else
      add = log(1+add);
    F += add;
  }
  return F;
}



double simpleHBLR::classify ( const Doc * doc ) {
  double sc = bias ? w[P-1] : 0;
  FLOOP ( f , *doc ) if ( *f.id < P-bias ) sc += w[*f.id] * (*f.val);
  return sc;
}


void simpleHBLR::writeBinary ( const char *fname ) {
  pthread_mutex_lock( &filelocks[label] );

  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  int K = 1;
  gzwrite( f , &K , sizeof(int) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , &bias , sizeof(int) );
  gzwrite( f , &label  , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);

  pthread_mutex_unlock( &filelocks[label] );
}


void simpleHBLR::readBinary ( int lockl , const char *fname ) {
  pthread_mutex_lock( &filelocks[lockl] );

  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  int K = 1;
  gzread( f , &K , sizeof(int) );
  gzread( f , &P , sizeof(int) );
  gzread( f , &bias , sizeof(int) );
  gzread( f , &label , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  pthread_mutex_unlock( &filelocks[lockl] );
}


simpleHBLR::~simpleHBLR (){
  free(w);
}

