#include <simpleHBLRAk.h>
#include <cstdlib>
#include <zlib.h>
#include <cstring>
#include <cmath>
#include <vector>
#include <lbfgssolver.h>
#include <cassert>
#include <simpleBLR.h>
#include <sys/time.h>

simpleHBLRAk::simpleHBLRAk ( int lockl , const char *fname , HInfo * _hinfo ) {
  dataptr = NULL, hinfo = _hinfo;
  param = new nodeParam(), parent = new nodeParam();
  param->readBinary( lockl , fname );
}


simpleHBLRAk::simpleHBLRAk ( Data* _dataptr , HInfo * _hinfo ) {
  dataptr = _dataptr;
  hinfo = _hinfo;
  param = new nodeParam();
  parent = new nodeParam();
}


void simpleHBLRAk::loadRequired( int label , const char *path ) {
  char fname[512];

  // Load current
  sprintf(fname,"%s/%d.gz", path , label );
  if ( fexist(fname) ) 
    param->readBinary(label,fname);
  else assert(0);

  // Load parent
  int par = hinfo->Parent(label);
  sprintf(fname,"%s/%d.gz", path , par );
  if ( fexist(fname) )
    parent->readBinary(par,fname);
  else if ( par != -1 ) assert(0);

  // Load children
  const std::vector<int>& c = hinfo->Children(label);
  for ( int i = 0; i < int(c.size()); ++i ) {
    sprintf(fname,"%s/%d.gz", path , c[i] );
    if ( fexist(fname) ) {
      nodeParam *child = new nodeParam();
      child->readBinary(c[i],fname);
      children.push_back( child );
    }
    else assert(0);
  }
}


double simpleHBLRAk::solve ( const char *path , int label , double eps ) {
  loadRequired( label , path );
  double F = 0;

  if ( hinfo->isLeaf( label ) == 1) 
    F = solveLeaf( eps );
  else
    F = solveNonLeaf();

  solveAlpha ( path , label );
  //  solvePrior ( path , label );
  return F;
}


void simpleHBLRAk::solveAlpha ( const char *path , int label ) {
  //loadRequired( label , path );

  double diff = 0;
  int n = std::max( param->P , parent->P);
  for ( int i = 0; i < n; ++i ) {
    double a = ( i < param->P ? param->w[i] : 0 );
    double b = ( i < parent->P ? parent->w[i] : 0 );
    diff += (a-b)*(a-b);
  }

  double denom = diff + param->sigma + parent->sigma;
  param->ap = ( param->aprior + .5*param->P);
  param->bp = ( param->bprior + .5*denom );

  param->alpha = (param->aprior + .5*param->P) / (param->bprior + .5*denom );
  //printf("%lf\n", param->alpha );
}

double digamma(double x);

void simpleHBLRAk::solvePrior ( const char * path , int label ) {
  /*
  double sumx = 0, sumlogx = 0, cnt = 0;

  for ( int i = 0; i < children.size(); ++i ) {
    sumx += children[i]->alpha;
    sumlogx += log( children[i]->alpha );
    cnt++;
  }
  
  double s = log(sumx/cnt) - sumlogx/cnt;
  if ( fabs(s-1) < 1e-2 || fabs(s) < 1e-2 || isnan(s) ) return;
  param->aprior = ( 3 - s + sqrt((s-3)*(s-3)+24*s) ) / (12*s );
  param->bprior = 1/param->aprior * sumx/cnt;
  printf("s = %lf , %lf %lf %lf\n", s , param->aprior , param->bprior , param->aprior / param->bprior);
  */
  assert( (param->ap/param->bp - param->alpha )  < 1e-9 );
  double s = (digamma(param->ap) + log(param->bp)) - log(param->alpha);
  param->aprior = ( 3 - s + sqrt((s-3)*(s-3)+24*s) ) / (12*s );
  param->bprior = 1/param->aprior * param->alpha;
  //  printf("%lf ", param->aprior / param->bprior );
}

#define INF (1<<30)
#define EPS (1e-50)

double simpleHBLRAk::solveLeaf ( double eps )  {
  assert( dataptr );

  param->P = std::max(param->P,parent->P);
  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) param->P = std::max(param->P,*f.id+1);
  }

  param->w = (double*) realloc(param->w,param->P*sizeof(double));

  memset( param->w , 0 , param->P*sizeof(double));

  LBFGSsolver* solver = new LBFGSsolver( param->w , param->P , this , std::min(1e-5,param->alpha*10) ) ;
  double F = solver->objective();
  delete solver;


  double *invSigma = (double*) malloc(param->P*sizeof(double));
  memset( invSigma , 0 , sizeof(double)*param->P);

  DLOOP ( doc , *dataptr ) {
    double sc = param->w[0];
    FLOOP ( f , *doc ) sc += param->w[*f.id] * (*f.val);
    sc = isinf(sc=exp(sc)) ? INF: sc;
    double pr = sc / ( 1 + sc );
    pr = ( pr < EPS ? EPS : pr > 1-EPS ? 1-EPS : pr );
    FLOOP ( f , *doc ) invSigma[*f.id] += pr*(1-pr)*(*f.val)*(*f.val);
  }

  param->sigma = 0;
  for ( int i = 1; i < param->P; ++i ) 
    param->sigma += 1.0/( invSigma[i] + param->alpha);

  free(invSigma);
  return F;
}


double simpleHBLRAk::solveNonLeaf () {
  param->P = std::max(param->P,parent->P);
  for ( int i = 0; i < int(children.size()); ++i ) {
    param->P = std::max(param->P,children[i]->P);
  }
  param->w = (double*) realloc(param->w,param->P*sizeof(double));
  memset( param->w , 0 , sizeof(double)*param->P);

  double sumalpha = 0;
  for ( int i = 0; i < parent->P; ++i ) 
    param->w[i] = param->alpha * parent->w[i];
  sumalpha += param->alpha;

  for ( int c = 0; c < int(children.size()); ++c ) {
    for ( int i = 0; i < children[c]->P; ++i ) 
      param->w[i] += children[c]->alpha * children[c]->w[i];
    sumalpha += children[c]->alpha;
  }

  double F = 0;
  for ( int i = 0; i < param->P; ++i )
    param->w[i] /= sumalpha;

  int n = std::max( param->P , parent->P );
  for ( int i = 0; i < n; ++i ) {
    double a = ( i < param->P ? param->w[i] : 0 );
    double b = ( i < parent->P ? parent->w[i] : 0 );
    F += param->alpha/2 * (a-b)*(a-b);			    
  }

  param->sigma = param->P/sumalpha;

  return F;
}



#define INF (1<<30)
#define EPS (1e-50)

double simpleHBLRAk::fg ( const double *x , double *g , int n ) {
  double F = 0;
  g[0] = 0;
  for ( int i = 1; i < param->P; ++i ) {
    double wp = (i < parent->P ? parent->w[i] : 0 );
    g[i] = param->alpha * (x[i]- wp);
    F += param->alpha/2 * (x[i]-wp)*(x[i]-wp);
  }

  DLOOP ( doc , *dataptr ) {
    double sc = x[0];
    FLOOP ( f , *doc ) sc += x[*f.id] * (*f.val);
    sc = isinf(sc=exp(sc)) ? INF: sc;

    double pr = sc / ( 1 + sc );
    pr = ( pr < EPS ? EPS : pr > 1-EPS ? 1-EPS : pr );
    int y = ( param->label == doc->labels[0] );
    FLOOP ( f , *doc ) g[*f.id] += -(y-pr) * (*f.val);
    g[0] += -(y-pr);
    F += - (y*log(pr) + (1-y)*log(1-pr)) ;
  }
  return F;
}


double simpleHBLRAk::classify ( const Doc * doc ) {
  double sc = param->w[0];
  FLOOP ( f , *doc ) if ( *f.id < param->P ) sc += param->w[*f.id] * (*f.val);
  sc = isinf(sc=exp(sc)) ? INF: sc;  
  return sc/(1+sc);
}


void simpleHBLRAk::writeBinary ( const char *fname ) {
  param->writeBinary(fname);
}

simpleHBLRAk::~simpleHBLRAk (){
  if ( param ) delete param;
  if ( parent ) delete parent;
  for ( int i = 0; i < int(children.size()); ++i ) 
    if ( children[i] ) delete children[i];
}





double digamma(double x)
{
  double result;
  static const double
    neginf = -1.0/0.0,
    c = 12,
    s = 1e-6,
    d1 = -0.57721566490153286,
    d2 = 1.6449340668482264365, /* pi^2/6 */
    s3 = 1./12,
    s4 = 1./120,
    s5 = 1./252,
    s6 = 1./240,
    s7 = 1./132,
    s8 = 691/32760,
    s9 = 1/12,
    s10 = 3617/8160;
  /* Illegal arguments */
  if((x == neginf) || isnan(x)) {
    return 0.0/0.0;
  }
  /* Singularities */
  if((x <= 0) && (floor(x) == x)) {
    return neginf;
  }
  /* Negative values */
  /* Use the reflection formula (Jeffrey 11.1.6):
   * digamma(-x) = digamma(x+1) + pi*cot(pi*x)
   *
   * This is related to the identity
   * digamma(-x) = digamma(x+1) - digamma(z) + digamma(1-z)
   * where z is the fractional part of x
   * For example:
   * digamma(-3.1) = 1/3.1 + 1/2.1 + 1/1.1 + 1/0.1 + digamma(1-0.1)
   *               = digamma(4.1) - digamma(0.1) + digamma(1-0.1)
   * Then we use
   * digamma(1-z) - digamma(z) = pi*cot(pi*z)
   */
  if(x < 0) {
    return digamma(1-x) + M_PI/tan(-M_PI*x);
  }
  /* Use Taylor series if argument <= S */
  if(x <= s) return d1 - 1/x + d2*x;
  /* Reduce to digamma(X + N) where (X + N) >= C */
  result = 0;
  while(x < c) {
    result -= 1/x;
    x++;
  }
  /* Use de Moivre's expansion if argument >= C */
  /* This expansion can be computed in Maple via asympt(Psi(x),x) */
  if(x >= c) {
    double r = 1/x;
    result += log(x) - 0.5*r;
    r *= r;
    result -= r * (s3 - r * (s4 - r * (s5 - r * (s6 - r * s7))));
  }
  return result;
}
