#include <simpleHDCD.h>
#include <cstdlib>
#include <zlib.h>
#include <cstring>
#include <cmath>
#include <vector>
#include <lbfgssolver.h>
#include <cassert>
#include <simpleBLR.h>
#include <sys/time.h>

simpleHDCD::simpleHDCD ( int _label , const char *fname , HInfo * _hinfo ) {
  dataptr = NULL, label = _label, w = NULL, wp = NULL, bias = 0, hinfo = _hinfo;
  lambda = 0;
  P = 0;
  readBinary( label , fname );
}


simpleHDCD::simpleHDCD ( Data* _dataptr , double _lambda , HInfo * _hinfo, int _bias ) {
  dataptr = _dataptr;
  hinfo = _hinfo;
  lambda = _lambda;
  label = -1;
  bias = _bias;
  wp = NULL;
  P = 0;

  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
  }
  P += bias;

  w = (double*) malloc(sizeof(double)*P);
  memset( w , 0 , sizeof(double)*P );
}


double simpleHDCD::solve ( const char *path , int _label , double eps ) {
  
  if ( hinfo->isLeaf( _label ) == 1) 
    return solveLeaf ( path , _label , eps );
  else
    return solveNonLeaf ( path , _label , eps );
}

double simpleHDCD::solveNonLeaf ( const char *path , int _label , double eps ) {
  label = _label;
  assert( P > 0 );
  double F = 0;
  int cnt = 0;
  char fname[512];
  std::vector<int> avgs = hinfo->Children(label);

  int debug = 0;
  if ( debug ) printf("label = %d\n", label );

  for ( int i = 0; i < avgs.size(); ++i ) {
    sprintf(fname,"%s/%d.gz", path , avgs[i] );
    if ( fexist(fname) ) {
      simpleHDCD N ( avgs[i] , fname , hinfo );
      for ( int j = 0; j < N.P; ++j )
	w[j] += N.w[j];
      if( debug ) { printf(" CHILD : "); N.print( stdout ); }
    }
    cnt++;
  }

  int par = hinfo->Parent(label);
  sprintf(fname,"%s/%d.gz", path , par );
  if ( fexist(fname) ) {
    // parent is non-zero
    cnt++;
    simpleHDCD N ( par , fname , hinfo );
    for ( int j = 0; j < N.P; ++j ) 
      w[j] += N.w[j];
    for ( int j = 0; j < P; ++j ) 
      w[j] /= cnt, F += .5*(w[j]-N.w[j])*(w[j]-N.w[j]);
    if( debug ) { printf(" PARENT : ");N.print( stdout );}
  } 
  else if ( par != -1 ) {
    // parent is zero
    cnt++;
    for ( int j = 0; j < P; ++j ) 
      w[j] /= cnt, F += .5*(w[j]*w[j]);
    if( debug ) printf(" PARENT : <zeros>");
  }
  else {
    // parent does not exist
    for ( int j = 0; j < P; ++j ) 
      w[j] /= (cnt+1), F += .5*(w[j]*w[j]);
    //debug = 1;
    if( debug ) printf(" PARENT : Does not Exist");
    debug = 0;
  }

  if( debug ) { printf("CURR: "); this->print( stdout );}
  F = 0;
  return F;
}





#define INF (1<<30)

double simpleHDCD::solveLeaf ( const char *path , int _label , double eps )  {
  assert( dataptr );
  label = _label;

  wp = (double*) realloc(wp,sizeof(double)*P);
  memset( wp , 0 , sizeof(double)*P);

  char fname[512];
  sprintf(fname,"%s/%d.gz", path , hinfo->Parent(label) );
  if ( fexist(fname) ) {
    simpleHDCD Par( hinfo->Parent(label) , fname , hinfo );
    memcpy( wp , Par.w , sizeof(double)*Par.P);
  }

  /* other opt , .5|w|^2 + .5|w-w_p|^2 instead of just .5|w-w_p|^2*/
  /*
  memcpy( w , wp , sizeof(double)*P);
  for ( int i = 0; i < P; ++i ) 
    w[i] /= 2;
  */

  assert( dataptr );
  label = _label;

  double F = 0;

  int N = dataptr->ndocs;

  int *index = new int[N], *target = new int[N],max_iter = 1000, active_size = N;
  double *QD = new double[N],*alpha = new double[N],PG,PGmax_old = INF, PGmin_old = -INF;
  double *pscore = new double[N];
  double C = lambda;


  // Init
  int npos = 0, nneg = 0;
  for ( int i = 0; i < N; ++i ) {
    alpha[i] = 0, QD[i] = 0, index[i] = i, pscore[i] = 0;
    FLOOP ( f , dataptr->docs[i] ) QD[i] += (*f.val)*(*f.val);
    if( bias ) QD[i] += 1; // 1 for the Bias term
    bool found = false;
    LLOOP ( l , dataptr->docs[i] ) found |= ( *l == label );

    target[i] = ( found == true ? +1 : -1 );
    npos += ( found ), nneg += (!found);
    FLOOP ( f , dataptr->docs[i] ) pscore[i] += (*f.val)*wp[*f.id];
    if( bias ) pscore[i] += wp[P-1];
    pscore[i] = target[i]*pscore[i]; 
  }
  assert( npos && nneg );

  // Perform Iterations

  for( int iter = 0; iter < max_iter;  ){
    double PGmax_new = -INF, PGmin_new = INF;
    for ( int i = 0; i < active_size; ++i ) std::swap( index[i] , index[ i+rand()%(active_size-i) ] );
    for ( int s = 0; s < active_size; ++s ) {
      int i = index[s];
      double G = 0;
      FLOOP ( f , dataptr->docs[i] ) G += w[ (*f.id) ] * (*f.val);
      if( bias ) G += w[P-1] * 1; // adding the bias term
      G = G * target[i] - (1-pscore[i]);
      PG = 0;

      if( alpha[i] == 0 ) {
	if( G > PGmax_old ) {
	  active_size--;
	  std::swap( index[s] , index[active_size] );
	  s--;
	  continue;
	}
	else if( G < 0 ) PG = G;
      } 
      else if( alpha[i] == C ) {
	if( G < PGmin_old ){
	  active_size--;
	  std::swap(index[s],index[active_size]);
	  s--;
	  continue;
	}
	else if( G > 0 ) PG = G;
      } 
      else PG = G;

      PGmax_new = std::max( PGmax_new , PG );
      PGmin_new = std::min( PGmin_new , PG );

      if( fabs(PG) > 1.0e-12 ) {
	double alpha_old = alpha[i];
	alpha[i] = std::min( std::max( alpha[i] - G/QD[i] , 0.0 ) , C );
	double d = (alpha[i] - alpha_old) * target[i];

	FLOOP(f,dataptr->docs[i]) w[ *f.id ] += d * (*f.val);
	if( bias ) w[P-1] += d * 1; // For the bias term
      }
    }
    iter++;
    if( PGmax_new - PGmin_new <= eps ) {
      if( active_size == N ) break;
      else {
	active_size = N;
	PGmax_old = INF;
	PGmin_old = -INF;
	continue;
      }
    }

    PGmax_old = PGmax_new;
    PGmin_old = PGmin_new;

    if( PGmax_old <= 0 ) PGmax_old = INF;
    if( PGmin_old >= 0 ) PGmin_old = -INF;
  }

  for ( int i = 0; i < P; ++i )
    w[i] += wp[i];

  F = 0;
  for ( int i = 0; i < P; ++i ) {
    F += .5*(w[i]-wp[i])*(w[i]-wp[i]);
  }

  for ( int i = 0; i < N; ++i ) {
    double score = classify( &dataptr->docs[i] );
    score = std::max( 1 - target[i]*score , 0.0 );
    F += C*score;
  }

  delete[] index;
  delete[] target;
  delete [] QD;
  delete[] alpha;
  delete[] pscore;



  free(wp),  wp = NULL;

  return F;
}






void simpleHDCD::print ( FILE * f) {

  int K = 1;
  fprintf(f,"{K:%d, P:%d, bias:%d, label:%d}\n", K , P , bias, label );
  fprintf(f,"{");
  for ( int i = 0; i < P; ++i ) 
    fprintf(f,"%d:%lf ", i , w[i] );
  fprintf(f,"}\n");

}




double simpleHDCD::classify ( const Doc * doc ) {
  double sc = bias ? w[P-1] : 0;
  FLOOP ( f , *doc ) if ( *f.id < P-bias ) sc += w[*f.id] * (*f.val);
  return sc;
}


void simpleHDCD::writeBinary ( const char *fname ) {
  pthread_mutex_lock( &filelocks[label] );

  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  int K = 1;
  gzwrite( f , &K , sizeof(int) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , &bias , sizeof(int) );
  gzwrite( f , &label  , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);

  pthread_mutex_unlock( &filelocks[label] );
}


void simpleHDCD::readBinary ( int lockl , const char *fname ) {
  pthread_mutex_lock( &filelocks[lockl] );

  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  int K = 1;
  gzread( f , &K , sizeof(int) );
  gzread( f , &P , sizeof(int) );
  gzread( f , &bias , sizeof(int) );
  gzread( f , &label , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  pthread_mutex_unlock( &filelocks[lockl] );
}


simpleHDCD::~simpleHDCD (){
  free(w);
}

