#include <simpleMLR.h>
#include <data.h>
#include <zlib.h>
#include <lbfgssolver.h>
#include <set>
#include <cmath>
#include <cassert>
#include <cstdlib>
#include <cstring>

simpleMLR::simpleMLR ( const char *fname ) {
  dataptr = NULL, labels = NULL, w = NULL, bias = 0;
  lambda = 0;
  K = 0, P = 0;
  readBinary( fname );
}


simpleMLR::simpleMLR ( Data* _dataptr , double _lambda , int _bias  ) {
  dataptr = _dataptr;
  lambda = _lambda;
  bias = _bias;
  K = 0, P = 0;
  
  std::set<int> unique_labels;
  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
    LLOOP ( lab , *doc ) unique_labels.insert(*lab);
  }

  labels = (int*) malloc(sizeof(int)*unique_labels.size());
  for ( std::set<int>::iterator it = unique_labels.begin();
	it != unique_labels.end(); ++it )
    labels[K++] = (*it);

  w = (double*) malloc(sizeof(double)*(K*P+bias*K));
  memset( w , 0 , sizeof(double)*(K*P+bias*K) );
  printf("Starting solver\n");
  LBFGSsolver* solver = new LBFGSsolver( w , K*P+bias*K , this );
  delete solver;
  
}

#define INF (1<<30)
#define EPS (1e-20)
double simpleMLR::fg ( const double *x , double *g , int n ) {

  double F = 0;
  for ( int i = 0; i < n-bias*K; ++i ) 
    g[i] = lambda * x[i],
    F += lambda/2 * x[i]*x[i];
  for ( int i = n-bias*K; i < n; ++i )
    g[i] = 0;

  double *sc = new double[K];

  DLOOP ( doc , *dataptr ) {
    double sum = 0;
    #pragma omp parallel for reduction(+: sum)
    for ( int k = 0; k < K; ++k ) {
      double s = bias ? x[K*P+k] : 0;
      FLOOP ( f , *doc ) s += x[k*P+(*f.id)] * (*f.val);
      s = isinf(s=exp(s)) ? INF: s;
      sum += s;
      sc[k] = s;
    }

    double Fadd = 0;
    #pragma omp parallel for reduction(+: Fadd)
    for ( int k = 0; k < K; ++k ) {
      double pr = sc[k] / sum;
      int y = ( labels[k] == doc->labels[0] );
      FLOOP ( f , *doc ) g[k*P+(*f.id)] += -(y-pr) * (*f.val);
      if ( bias ) g[K*P+k] += -(y-pr);

      pr = ( pr < EPS ? EPS : pr > 1-EPS ? 1-EPS : pr );
      Fadd += - y * log ( pr );
    }
    F += Fadd;
  }

  delete[] sc;
  //  printf("%lf\n", F );
  return F;
}





void simpleMLR::writeBinary ( const char *fname ) {
  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }
  gzwrite( f , &K , sizeof(K) );
  gzwrite( f , &P , sizeof(P) );
  gzwrite( f , labels , sizeof(int)*K );
  gzwrite( f , &bias , sizeof(bias) );
  gzwrite( f , w , sizeof(double)*(K*P+bias*K) );
  gzclose(f);
}

void simpleMLR::writeBinaryFolder ( const char *folder ) {

  char fname[512];
  for ( int k = 0; k < K; ++k ) {
    sprintf(fname,"%s/%d.gz",folder,labels[k]);

    gzFile f = gzopen( fname, "wb" );
    if ( f == NULL ) {
      perror("Cannot open data file for writing");
      exit(0);
    }
    int one = 1, wp = P+bias;
    gzwrite( f , &one , sizeof(one) );
    gzwrite( f , &wp , sizeof(wp) );
    gzwrite( f , &bias , sizeof(int) );
    gzwrite( f , &labels[k] , sizeof(int) );
    gzwrite( f , w+k*P , sizeof(double)*P );
    if ( bias ) gzwrite( f , w+K*P+k , sizeof(double) );
    gzclose(f);
  }
}



void simpleMLR::readBinary ( const char *fname ) {
  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  gzread( f , &K , sizeof(K) );
  gzread( f , &P , sizeof(P) );
  labels = (int*) realloc(labels,sizeof(int)*K);
  gzread( f , labels , sizeof(int)*K );
  gzread( f , &bias , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*(K*P+bias*K));
  gzread( f , w , sizeof(double)*(K*P+bias*K) );
  gzclose(f);
}


simpleMLR::~simpleMLR (){
  if (labels) free(labels);
  if (w) free(w);
}

